# $Id: config.py,v 1.5 2004/03/17 11:36:18 christht Exp $
"""Module contains main function 'parse(filename=None)' for parsing
config for swup. Function returns a dictionary of config variables. If
a valid optional filename is given, variables defined in file will
override the corresponding default variables.
"""

#  Copyright 2001 - 2003 Trustix AS - <http://www.trustix.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


import shlex, os, string, sys
from swuplib import log

false, true = 0, 1
tell = 1

# defaults
DEFAULT_CONFIG_FILE = "/etc/swup/swup.conf"
DEFAULT_CONFIG = {
"driver": "rpmdriver",
"gnupgdir": "/etc/swup/gnupg",
"cachedir": "/var/cache/swup",
"interactive_logged": "yes",
"download_only": "no",
"save_to":  "",
"poll_only": "no",
"tmpdir": "/var/spool/swup",
"sites": None,
"local_first": "no",
"exclude_pkg_regexp": "",
"include_pkg_regexp": ".*",
}


class ConfigError(Exception): pass
class ConfigIOError(ConfigError): pass

class ConfigParseError(ConfigError):
    def __init__(self, filename, config, msg=None):
        self.file = filename
        self.lineno = config.lineno
        self.token = config.token
        self.msg = msg

    def __str__(self):
        msg = 'Error in ' + self.file + ', line ' + str(self.lineno)
        if self.token != "":
            msg = msg + " near '" + self.token + "'"
        msg = msg + ": " + self.msg
        return msg

class ConfigBadOptionError(ConfigError):
    def __init__(self, option):
        self.option = option

    def __str__(self):
        return 'Unrecognized option ' + self.option

class ConfigDeprecatedOptionError(ConfigError):
    def __init__(self, option, new):
        self.option = option
	self.new = new

    def __str__(self):
        return "Deprecated option '" + self.option \
		+ "' should be replaced by '" \
		+ self.new + "'."

class ConfigDeprecatedOptionRemovedError(ConfigError):
    def __init__(self, option ):
        self.option = option

    def __str__(self):
        return "Deprecated option '" \
		+ self.option + "' ignored."

class ConfigMissingOptionError(ConfigError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg
        
class ConfigNoNullClassFoundError(ConfigError):
    def __init__(self):
        self.msg = "Error: No site configured with 'class = 0' found."

    def __str__(self):
        return self.msg

class ConfigFewNullClassFoundError(ConfigError):
    def __init__(self):
        self.msg = "Only one site configured with 'class = 0' found."

    def __str__(self):
        return self.msg

class ConfigNoSiteFoundError(ConfigError):
    def __init__(self):
        self.msg = "Error: No configured sites found."

    def __str__(self):
        return self.msg

class ConfigSection:
    def __init__(self, type, log):
        self.type = type
	self.log = log
        self.options = {}

    def __str__(self):
        return str(self.options)

    def set(self, option, value):
        if option == "priority":
	    new = "class"
	    errmsg = ConfigDeprecatedOptionError( option ,new)
            self.log.write_syslog_warning( "%s" %errmsg )
            if tell == 1:
                self.log.write_stdout( errmsg, true )
            option = new
	if option == "class":
            self.options[option] = int(value)
        elif option in ["regexp", "location"]:
            self.options[option] = str(value)
        else:
            raise ConfigBadOptionError( option )

    def getType(self):
        return self.type

    def validate(self):
        """This function makes sure that all required options are set."""
        keys = self.options.keys()
        if not 'location' in keys:
            raise ConfigMissingOptionError("Missing URL for site.")
        if not 'regexp' in keys:
            self.set("regexp", ".*")
        if not 'class' in keys:
            self.set("class", 0)
            
    def __len__(self):
        return len(self.options)

    def __getitem__(self, key):
        return self.options[key]

    def has_key(self, key):
        return self.options.has_key(key)

    def values(self):
        return self.options.values()

    def __delitem__(self, key):
        del self.options[key]

    def update(self, dict):
        self.options.update(dict)


class ConfigParser:
    def __init__( self , tell, log ):
        "Set up configuration and run initialization functions."
        self.log = log
        self.options = {}
        self.sections = []
        self.tell = tell

    def stripQuotes(self, text):
        if len(text) > 1:
            if text[0] == '"' and text[-1] == '"' \
               or text[0] == "'" and text[-1] == "'":
                return text[1:-1]
        return text
    
    
    def addOption(self, option, value):
	# Check for deprecated options:
	error = 0
        if option == "passwd_file":
		errmsg = ConfigDeprecatedOptionRemovedError(option)
                self.log.write_syslog_warning( "%s" %errmsg )
                if self.tell == 1:
                    self.log.write_stdout( "Warning: %s" %errmsg, true )
                error = 1
	if option == "logfile":
		errmsg = ConfigDeprecatedOptionRemovedError(option)
                self.log.write_syslog_warning( "%s" %errmsg )
                if self.tell == 1:
                    self.log.write_stdout( "Warning: %s" %errmsg, true )
                error = 1
	if option == "flush_cache":
		errmsg = ConfigDeprecatedOptionRemovedError(option)
                self.log.write_syslog_warning( "%s" %errmsg )
                if self.tell == 1:
                    self.log.write_stdout( "Warning: %s" %errmsg, true )
                error = 1
	if option == "exclude_group_regexp":
		errmsg = ConfigDeprecatedOptionRemovedError(option)
                self.log.write_syslog_warning( "%s" %errmsg )
                if self.tell == 1:
                    self.log.write_stdout( "Warning: %s" %errmsg, true )
                error = 1
	if option == "include_group_regexp":
		errmsg = ConfigDeprecatedOptionRemovedError(option)
                self.log.write_syslog_warning( "%s" %errmsg )
                if self.tell == 1:
                    self.log.write_stdout( "Warning: %s" %errmsg, true )
                error = 1
        if not error == 0:
                return


        # Validate type and possibly raise exception if
        # the option is unknown.
        elif option not in DEFAULT_CONFIG.keys():
            raise ConfigBadOptionError(option)
        else:
            self.options[option] = self.stripQuotes(value)
    
    
    def parse(self, filename):
        try:
            fd = open(filename)
        except IOError:
            raise ConfigIOError(filename)
        except:
            raise
        config = shlex.shlex(fd)
        finished = 0
        state = '-'
        option = None
        section = None
        while not finished:
            token = config.get_token()
            if token == "":
                finished = 1
            else:
                if state == '-':
                    # reading option name
                    option = token
                    state = 'O'

                elif state == 'O':
                    # next token should be '=' for assignment, or
                    # '{' to start a section
                    if token == '=':
                        state = '='
                    elif token == '{':
                        if option != "site":
                            raise ConfigParseError(filename, config, "Unrecognized section " + option)
                        state = '{'
                        section = ConfigSection(option, self.log)
                    else:
                        raise ConfigParseError(filename, config, "Expecting '=' or '{'")

                elif state == '=':
                    # next token is the value of the option
                    try:
                        self.addOption( option, token )
                        option = None
                    except ConfigBadOptionError, e:
                        raise ConfigParseError(filename, config, e.__str__())
                    except:
                        raise
                    state = '-'
                
                elif state == '{':
                    # next token should be an option in the section,
                    # or '}' to finish the section
                    if token == '}':
                        try:
                            section.validate()
                            self.sections.append(section)
                            section = None
                        except ConfigMissingOptionError, e:
                            raise ConfigParseError(filename, config, e.__str__())
                        except:
                            raise
                        state = '-'
                    else:
                        option = token
                        state = '{O'

                elif state == '{O':
                    # next token should be '=' for assignment
                    if token == '=':
                        state = '{='
                    else:
                        raise ConfigParseError(filename, config, "Expecting '='")

                elif state == '{=':
                    # next token is the value of the option
                    try:
                        section.set(option, self.stripQuotes(token))
                    except ConfigBadOptionError, e:
                        raise ConfigParseError(filename, config, e.__str__())
                    except:
                        raise
                    option = None
                    state = '{'
        fd.close()
        
    def getSections(self):
        return self.sections

    def getOptions(self):
        return self.options

def parse(filename=None, silent=None):
    if silent == 1:
        tell = 0
    else:
        tell = 1

    config = DEFAULT_CONFIG.copy()
    if filename:
        if not os.path.isfile(filename):
            raise ConfigIOError(filename)
        logger = log.log(log.LOGLEVEL_NORMAL, true )
        parser = ConfigParser( tell, logger )
        parser.parse(filename)
        options = parser.getOptions()
        config.update(options)
        sections = parser.getSections()
        sitelist = []
        for section in sections:
            if section.getType() == 'site':
                sitelist.append((section['class'], section['location'], section['regexp']))
        if sitelist:
            nulls = 0
            for site in sitelist:
                (pri, url, regexp) = site
                if pri == 0:
                    nulls += 1
            if nulls == 0:
                raise ConfigNoNullClassFoundError()
            elif nulls == 1:
                errmsg = ConfigFewNullClassFoundError()
                sys.stdout.write( "Warning: %s\n" %errmsg)
            config["sites"] = sitelist
        else:
            raise ConfigNoSiteFoundError()

    # check uid != 0
    if os.getuid() != 0:
        cachedir= os.path.join(os.path.expanduser("~"), '.swup/cache')
        tmpdir= os.path.join(os.path.expanduser("~"), '.swup/tmp')
        gnupgdir= os.path.join(os.path.expanduser("~"), '.swup/gnupg')
        config['cachedir'] = cachedir
        config['tmpdir'] = cachedir
        config['gnupgdir'] = cachedir
    # set up listsdir and md5sumfile
    config['listsdir'] = os.path.join(config['cachedir'], 'lists')
    config['md5sumfile'] = os.path.join(config['cachedir'], 'md5sums')
    return config




