# -*- python -*-
# $Id: swup.py,v 1.30 2004/03/17 11:36:18 christht Exp $

#  Copyright 2001 - 2003 Trustix AS - <http://www.trustix.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


import sys, string, os, re
from swuplib import cmdline
from swuplib import config
from swuplib import log
from swuplib import help

# Exceptions
class SwupException(Exception):
    def __init__(self, value=None):
        self.value = value
    def __str__(self):
        return self.value

class ConfigException(SwupException): pass
class CmdlineException(SwupException): pass
class SetupException(SwupException): pass
class ImportKeyException(SwupException): pass
class DeleteKeyException(SwupException): pass
class ListKeysException(SwupException): pass


def parse_argv(argv):
    '''Parse command line arguments. Return dictionary with switch as key.'''
    try:
        argdict = cmdline.parse(argv)
    except cmdline.unknown, option:
        raise CmdlineException, "Unknown option: %s" % option
    except cmdline.wrong_number_arguments, switch:
        raise CmdlineException, "Wrong number of arguments for switch: %s"\
              % switch
    except cmdline.too_many_operations, (operation1, operation2):
        raise CmdlineException, "Too many operations: %s %s" %\
              (operation1, operation2)
    except cmdline.too_few_operations, message:
        raise CmdlineException, "Too few operations: %s" % message
    except:
        raise
    return argdict


def import_key(filename, gnupgdir):
    '''Given a local filename that contains a gnupg key and the dictionary
    in which the keyring is stored, import key into keyring.'''
    if os.path.isfile(filename):
        try:
            from swuplib import gpg
            (retval, output) = gpg.import_key(filename, gnupgdir)
            sys.stdout.write("Import of key(s) succeded.\n")
        except gpg.key_import_failure, errmsg:
            raise ImportKeyException, errmsg
        except:
            raise
    else:
        raise ImportKeyException, "No such file '%s'" % filename

def delete_key(keyid, gnupgdir):
    '''Deletes key from keyring.'''
    try:
        from swuplib import gpg
    	(retval, output) = gpg.delete_key(keyid, gnupgdir)
    except gpg.key_delete_failure, errmsg:
    	raise DeleteKeyException, errmsg
    except:
        raise

def list_keys(gnupgdir):
    '''Given a directory where the gnupg keyring is stored, list the keys
    in the keyring.'''
    try:
        from swuplib import gpg
        (retval, output) = gpg.list_keys(gnupgdir)
        sys.stdout.write(output)
    except gpg.list_keys_failure, errmsg:
        raise ListKeysException, errmsg
    except:
        raise

def lock_ignored(config):
    if config.has_key('ignore_lock') and config['ignore_lock'] == "yes":
        return True
    return False

def check_lock(config):
    if os.path.exists(os.path.join(config['cachedir'],'.lock')):
        return False
    return True

def set_lock(config):
    if lock_ignored(config):
        return True
    cachedir = config['cachedir']
    lockfile = os.path.join(os.path.join(cachedir,'.lock'))
    if check_lock(config):
        if not os.path.isdir(cachedir):
            os.makedirs(cachedir)
        fd = open(lockfile, 'w')
        fd.write('1')
        fd.close()
        return True
    else:
        message = \
"Error: Lockfile '%s' exists. Exiting.\n\n" %lockfile +\
"This generally means that you have another swup process running, or that\n" +\
"the last swup process you ran died unexpectedly. The file may be manually\n"+\
"removed, but only do this if you are sure that you are not running any\n" +\
"other swup processeses.\n"
        sys.stderr.write(message)

    return False

def remove_lock(config):
    if lock_ignored(config):
        return True
    lockfile = os.path.join(os.path.join(config['cachedir'],'.lock'))
    if os.path.exists(lockfile):
        try:
            os.remove(lockfile)
        except IOError:
            return False
        except:
            raise
    return True
    
    
def setup(argdict):
    '''Analyze argument dictionary, return (mode, args, config)
    where mode is a string, args is a list of strings and config is
    a dictionary of configuration variables with variable name as key.'''

    # did user request help or version info?
    if argdict.has_key('--help'):
        return ('help', None, None)
    if argdict.has_key('--version'):
        return ('version', None, None)
    if argdict.has_key('--copyright'):
        return ('copyright', None, None)

    elif argdict.has_key('--config-file'):
	config_file = argdict['--config-file'][0]
    else:
	config_file = config.DEFAULT_CONFIG_FILE

    silent = 0
    if argdict.has_key('--silent'):
	silent = 1
    configuration = None
    try:
	configuration = config.parse(config_file, silent)
    except config.ConfigIOError, e:
        raise SetupException, 'Unable to open config file: %s' % e
    except config.ConfigNoNullClassFoundError, e:
        raise SetupException, str(e)
    except config.ConfigNoSiteFoundError, e:
        raise SetupException, str(e)
    except config.ConfigParseError, message:
        raise SetupException, '%s\n%s' % \
              ('Error in config file %s.' % config_file,
               'Parser said %s.' % message)
    except:
        raise

    # merge command line options with config
    if argdict.has_key('--silent'):
	configuration['loglevel'] = log.LOGLEVEL_SILENT 
    elif argdict.has_key('--quiet'):
        configuration['loglevel'] = log.LOGLEVEL_QUIET
    elif argdict.has_key('--verbose'):
        configuration['loglevel'] = log.LOGLEVEL_VERBOSE
    else:
        configuration['loglevel'] = log.LOGLEVEL_NORMAL
        
    if argdict.has_key('--repository-URI'):
	configuration['sites'] = [('0', argdict['--repository-URI'][0],
				       '.*')]
    if argdict.has_key('--gnupg-dir'):
	configuration['gnupgdir'] = argdict['--gnupg-dir'][0]
    if argdict.has_key('--save-to'):
        todir = argdict['--save-to'][0]
        if not os.path.isdir( todir ):
            raise SetupException, "No such directory '%s'" % todir
        else:
            configuration['save_to'] = todir
        
    if argdict.has_key('--import-key'):
	filename = argdict['--import-key'][0]
        return ('import_key', filename, configuration)

    if argdict.has_key('--delete-key'):
	keyid = argdict['--delete-key'][0]
        return ('delete_key', keyid, configuration)


    if argdict.has_key('--list-keys'):
        return ('list_keys', None, configuration)

    if argdict.has_key('--ignore-lock'):
        configuration['ignore_lock'] = 'yes'
    else:
        configuration['ignore_lock'] = 'no'

    if argdict.has_key('--ignore-filter'):
        configuration['ignore_filter'] = 'yes'
    else:
        configuration['ignore_filter'] = 'no'

    poll_only = False
    download_only = False

    configuration['no_swupgrade'] = False
    if argdict.has_key('--poll-only'):
	configuration['poll_only'] = 'yes'
        configuration['no_swupgrade'] = True
    if argdict.has_key('--download-only'):
	configuration['download_only'] = 'yes'
        configuration['no_swupgrade'] = True

    packages = []

    if argdict.has_key('--install'):
	packages.extend(argdict['--install'])
    if argdict.has_key('--upgrade'):
	packages.extend(argdict['--upgrade'])
    if argdict.has_key('--package'):
        if len(packages) < 1:
	    packages.extend(argdict['--package'])
    if argdict.has_key('--package-file'):
	try:
	    package_file = open(argdict['--package-file'][0])
	    package_file_list = package_file.readlines()
	    #package_file_list.remove('')
	    packages.extend(package_file_list)
	    package_file.close()
	except IOError: 
            raise SetupException, "Unable to open file: %s" %\
                  argdict['--package-file'][0]
        except:
            raise

    if argdict.has_key('--package-URI'):
	packages.extend(argdict['--package-URI'])

    if argdict.has_key('--package-URI-file'):
	try:
	    package_uri_file = \
			     open(argdict['--package-URI-file'][0])
	    package_uri_file_list = package_uri_file.readlines()
	    #package_uri_file_list.remove("")
	    packages.extend(package_uri_file_list)
	    package_uri_file.close()
	except IOError: 
            raise SetupException, "Unable to open URI-file: %s" %\
                  argdict['--package-file'][0]
        except:
            raise

    # prettify list
    for i in range(len(packages)):
	packages[i] = string.strip(packages[i])

    # create the installation object

    if argdict.has_key('--upgrade'):
	if packages == []:
            mode = 'upgrade_normal'
	else:
            mode = 'upgrade_package'
        return (mode, packages, configuration)

    elif argdict.has_key('--install'):
	if packages == []:
            raise SetupException, \
                  'You need to specify at least one package to install.'
        return ('install', packages, configuration)

    elif argdict.has_key('--search-package'):
        searchpattern = argdict['--search-package'][0]
        return ('search_package', searchpattern, configuration)

    elif argdict.has_key('--search-file'):
        searchpattern = argdict['--search-file'][0]
        return ('search_file', searchpattern, configuration)
    
    elif argdict.has_key('--list-latest'):
        return ('list_latest', None, configuration)

    # Sometime in the future this should go away
    elif argdict.has_key('--list-new'):
        return ('list_prospect', None, configuration)
        
    elif argdict.has_key('--list-prospect'):
        return ('list_prospect', None, configuration)
        
    elif argdict.has_key('--list-alien'):
        return ('list_alien', None, configuration)
        
    elif argdict.has_key('--list-upgrade'):
        return ('list_upgrade', None, configuration)
        
    elif argdict.has_key('--list-downgrade'):
        return ('list_downgrade', None, configuration)
    
    elif argdict.has_key('--flush-cache'):
	return ('flush_cache', None, configuration)

    elif argdict.has_key('--remove-lock'):
	return ('remove_lock', None, configuration)

    elif argdict.has_key('--describe'):
	configuration['descriptor'] = argdict['--describe'][0]
	return ('describe', None, configuration)

    else:
        raise SetupExeption, "This should not happen."


def main(argv):
    'Main interface. Takes list of command line arguments as argument.'
    swupgrade = False

    # Parse the command line arguments.
    try:
        argdict = parse_argv(argv)
    except CmdlineException, errmsg:
        sys.stderr.write('%s\n' % errmsg)
        sys.exit(2)
    except:
        raise

    # Setup the configuration and get the mode in which we are supposed to
    # run.
    try:
        mode, args, config = setup(argdict)
    except SetupException, errmsg:
        sys.stderr.write('%s\n' % errmsg)
        sys.exit(2)
    except:
        raise

    # Check for the simpler modes that does not involve the upgrade module.
    if mode == 'help':
        help.helpme()
        sys.exit(0)
    elif mode == 'version':
        help.version()
        sys.exit(0)
    elif mode == 'copyright':
        help.copyright()
        sys.exit(0)
    elif mode == 'remove_lock':
        remove_lock(config)
        sys.exit(0)
    elif mode == 'import_key':
        if set_lock(config):
            try:
                import_key(args, config['gnupgdir'])
                remove_lock(config)
                sys.exit(0)
            except ImportKeyException, errmsg:
                sys.stderr.write('%s\n' % errmsg)
                remove_lock(config)
                sys.exit(2)
            except:
                remove_lock(config)
                raise
    elif mode == 'delete_key':
        if set_lock(config):
            try:
                delete_key(args, config['gnupgdir'])
                remove_lock(config)
                sys.exit(0)
            except DeleteKeyException, errmsg:
                sys.stderr.write('%s\n' % errmsg)
                remove_lock(config)
                sys.exit(2)
            except:
                remove_lock(config)
                raise
    elif mode == 'list_keys':
        try:
            list_keys(config['gnupgdir'])
            sys.exit(0)
        except ListKeysException, errmsg:
            sys.stderr.write('%s\n' % errmsg)
            sys.exit(1)
        except:
            raise

    # The mode requires the upgrade module. Initialize an upgrade agent
    # and run.

    import upgrade
    from swuplib import download
    agent = None
    try:
        if mode == 'upgrade_normal':
            if not set_lock(config):
                sys.exit(1)
            agent = upgrade.Upgrade(config)
        elif mode == 'upgrade_package':
            if not set_lock(config):
                sys.exit(1)
            config['local_first'] = "yes"
            agent = upgrade.upgrade_package(args, config)
        elif mode == 'install':
            if not set_lock(config):
                sys.exit(1)
            agent = upgrade.install_package(args, config)
        elif mode == 'list_latest':
            agent = upgrade.list_latest(config)
        elif mode == 'list_prospect':
            agent = upgrade.list_prospect(config)
        elif mode == 'list_alien':
            agent = upgrade.list_alien(config)
        elif mode == 'list_upgrade':
            agent = upgrade.list_upgrade(config)
        elif mode == 'list_downgrade':
            agent = upgrade.list_downgrade(config)
        elif mode == 'search_package':
            agent = upgrade.search_package(args, config)
        elif mode == 'search_file':
            agent = upgrade.search_file(args, config)
	elif mode == 'flush_cache':
            if not set_lock(config):
                sys.exit(1)
	    agent = download.Flush( config )
	elif mode == 'describe':
	    agent = upgrade.describe( config )
    except upgrade.UpgradeException, err:
        sys.stderr.write('Failed to initialize: %s\n' % err)
        sys.exit(1)
    except:
        raise
    
    try:
        if agent:
            swupgrade = agent.run()
            remove_lock(config)
            if swupgrade:
                #
                # Swup was upgraded, most likely with close friends only.
                # Let's flush cache before we restart swup:
                #
                agent = download.Flush( config )
                agent.run()
        else:
            sys.stderr.write('This should never happen.\n')
    except KeyboardInterrupt:
        remove_lock(config)
	raise KeyboardInterrupt
    except upgrade.UpgradeException:
        remove_lock(config)
        sys.exit(1)
    except:
        remove_lock(config)
        raise

    # All went ok if we got here. Exit normally.
    if swupgrade and not config['no_swupgrade']:
        return swupgrade
    else:
        sys.exit(0)


if __name__ == '__main__':
    main(sys.argv[1:])

