# -*- python -*-
# $Id: upgrade.py,v 1.79 2004/03/17 15:09:35 christht Exp $

#  Copyright 2001 - 2003 Trustix AS - <http://www.trustix.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

DEBUG = True

INSTALL_FLAG = "install"
UPGRADE_FLAG = "upgrade"

DEPENDENCY_ERROR_MESSAGE = \
"""There were some problems with dependencies when upgrading or
installing.  Usually problems arise when an upgrade will break already
installed software. In order to upgrade the selected package(s), you
must also upgrade the already installed (and depending) packages. A
'swup --upgrade' will usually fix this."""

NOSITES_ERROR_MESSAGE = \
"""Unable to open a site or get valid upgrade information. This will
happen if all top priority sites are down, or if upgrade information
is not available or is missing a valid signature from a known public
key. Check site, configuration and public keyring."""

# exceptions
class UpgradeException(Exception):
    def __init__(self, value=None):
        self.value = value
    def __str__(self):
        return str(self.value)

class DownloadException(UpgradeException): pass
class ResolveException(UpgradeException): pass
class FlushCacheException(UpgradeException): pass
class ParseException(UpgradeException): pass
class AuthException(DownloadException): pass

NoValidSiteFoundError = "No site with valid filelist found."


import sys, re, os, string
import resolver
from swuplib import spiparser, download, log, utils
from swuplib.driver import pkgdriver
from UserDict import UserDict

    

class Upgrade:
    "Base class for upgrade."

    def _init_log(self):
        "Initialize log."
        if string.lower(self.config["interactive_logged"]) == "yes":
            interactive_logged = True
        else:
            interactive_logged = False
        
        self.log = log.log(self.config["loglevel"], interactive_logged )

    def _init_driver(self):
        "Initialize driver."
        self.pkgdrv = pkgdriver.pkgdriver(self.config, self.log)

    def _init_reader(self):
        "Initialize RDF reader."
        self.reader = spiparser.rdf_reader()


    def _init_downloader(self):
        "Initialize downloader."
        try:
            self._downloader = download.download(
                self.config["gnupgdir"],
                self.log,
                self.config["cachedir"],
                self.config["tmpdir"])
        except download.DownloadException, err:
            raise UpgradeException, err
        except:
            raise


    def _init_regexps(self):
        "Initialize regular expressions for filter."
        regexps = self._compile_regexps( [self.config["exclude_pkg_regexp"],
                                     self.config["include_pkg_regexp"]])
        (self.exclude_pkg_regexp,
         self.include_pkg_regexp) = regexps

        
    def init(self):
        "Run initialization functions."
        self._init_log()
        self._init_driver()
        self._init_reader()
        self._init_regexps()
        self._init_downloader()
        self.latest = {}
        self.obsoletes = {}
        self.obsoletors = {}
        self.emptylist = False
        

    def __init__(self, configdict=None):
        "Set up configuration and run initialization functions."
        if not configdict:
            from swuplib import config
            self.config = config.parse()
        else:
            self.config = configdict
        self.init()
        self.latest = {}
        self.obsoletes = {}
        self.obsoletors = {}
    
    
    def _compile_regexp(self, regexp):
        if regexp in [None, "", "None"]:
            return re.compile(".(?!.)")
        else:
            return re.compile(regexp)

    
    def _compile_regexps(self, regexps):
        """Compile list of strings into regexp objects. Returns a list of
        re.RegexObject instances."""
        retlist = []
        for regexp in regexps:
            retlist.append(self._compile_regexp(regexp))
        return retlist


    def _sort_2d_sitelist(self, element1, element2):
        pri1, pri2 = abs(int(element1[0])), abs(int(element2[0]))
        if  pri1 < pri2:
            return -1
        elif pri1 > pri2:
            return 1
        elif pri1 == pri2:
            return 0
        else:
            raise ValueError, "unable to compare"


    def load_data(self, filename):
        'Load a serialized object from file. Return object.'
        import cPickle
        fd = open(filename)
        retval = cPickle.load(fd)
        fd.close()
        return retval


    def save_data(self, obj, filename):
        'Serialize and save a object obj to file given by filename.'
        import cPickle
        dirname = os.path.dirname(filename)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        fd = open(filename, 'w')
        cPickle.dump(obj, fd)
        fd.close()


    def retrieve_latest(self, uri):
        """Parse package list rdf, given the uri to the rdf-file.  Returns
        a dictionary with package name as key, and then a dictionary of
        name, version, release, summary, uri and obsoletes as value for
        keys."""
        # Check md5sum, pickle.load and/or dump -- to be implemented.

        # Get the md5sum uri:
        if uri[-3:] == '.gz':
            md5sum_uri = uri[:-3] + '.md5'
            noncompressed_uri = uri[:-3]
        else:
            noncompressed_uri = uri
            md5sum_uri = uri + '.md5'

        dict = None

        # Check if md5sum has changed. The check will update md5sums
        # if they are changed.
        if self.md5sum_isequal(md5sum_uri, noncompressed_uri):
            # Pickle already parsed latest-information.
            pickle_file = os.path.join(self.config['listsdir'], uri)
            pickle_file = utils.normalize_uri(pickle_file + '.pickle')
            if os.path.isfile(pickle_file):
                dict = self.load_data(pickle_file)
                
        if not dict:
            #
            # Now we should remove rdfs for this site, as we need new ones:
            # Lets use the Flush class for this:
            #
            baseuri, basename = os.path.split(uri)
            flusher = download.Flush(self.config)
            flusher.run(baseuri)

            # Get latest, parse, and pickle-dump.
            #try to download gzipped file first
            try:
                local_filename = self.download_file(uri + '.gz', TYPE = 2)
            except IOError, DownloadException:
                local_filename = self.download_file(uri, TYPE = 2)
            except:
                raise
            
            self.log.write_tty('\nParsing file: %s -' % basename)
            dict = self.reader.read_latest( local_filename, baseuri )
            self.log.write_tty('\rParsing file: %s - done\n' % basename)
            pickle_file = os.path.join(self.config['listsdir'], uri)
            pickle_file = utils.normalize_uri(pickle_file + '.pickle')
            self.save_data(dict, pickle_file)
            self.md5sum_set(md5sum_uri,noncompressed_uri)
            #
            # This makes the downloader behave:
            os.remove(local_filename)
            
        return dict


    def _find_who_obsoletes( self, latestdict ):
        """Given a dictionary over latest packages, return a dictionary
        with obsoleted package as key and a list of packages that obsoletes
        the package as value for key.
        Used by _generate_latest_lists
        """
        obsoleted_by = {}
        for name in latestdict.keys():
            if latestdict[name]["obsoletes"]:
                for obs in latestdict[name]["obsoletes"]:
                    if not obsoleted_by.has_key(obs):
                        obsoleted_by[obs] = [name]
                    else:
                        obsoleted_by[obs].append(name)
        return obsoleted_by


    def _find_the_obsoletors( self, latestdict ):
        """Given a dictionary over latest packages, return a dictionary
        with packages that obsoletes others as key and a list of obsoletes
        packages the package as value for key.
        Used by _generate_latest_lists"""
        obsoletors = {}
        for name in latestdict.keys():
            if latestdict[name]["obsoletes"]:
                for obs in latestdict[name]["obsoletes"]:
                    if not obsoletors.has_key(name):
                        obsoletors[name] = [obs]
                    else:
                        obsoletors[name].append(obs)
        return obsoletors

    def _latest_merge( self, main_latest, latest):
        '''Given two dictionaries, merge the packages into the first, and 
        return it'''
        for name in latest.keys():
            if not main_latest.has_key(name):
                main_latest[name] = latest[name]
                continue
            result = self.pkgdrv.version_compare(main_latest[name],latest[name])
            if result < 0:
                main_latest[name] = latest[name]
        
        return main_latest

    def _generate_latest_lists(self):
        '''Updates the latest, obsoletes and obsoletors dicts if needed.'''
        #
        # We only check latest. Maybe there ar noe obsoletes or obsoletors at
        # all.
        if not self.latest:

            self.latest = {}
            self.obsoletes = {}
            self.obsoletors = {}
        
            #
            # Get list of sites:
            #
            sitelist = self.config["sites"]
            #
            # place sites in a list, and the list in a dict with class as key:
            #
            classlist = {}
            for site in sitelist:
                (pri, latest_uri, regexp) = site
                if not classlist.has_key(pri):
                    classlist[pri] = []
                classlist[pri].append((latest_uri, regexp))

            #
            # Now merge all sites belonging to one class into one latest dict:
            #
            latestlist = []
            prilist = classlist.keys()
            prilist.sort()
            for pri in prilist:
                #
                # Get latest
                #
                sitelist = classlist[pri]
                
                found_valid_site = 0
                valid_sitelist = []
                main_latest = {}
                for site in sitelist:
                    (latest_uri, regexp) = site
                    filter = self._compile_regexp(regexp)
                    if not filter:
                        continue
                    try:
                        message = "Fetching upgrade info '%s'" % latest_uri
                        self.log.write_syslog_info( message )
                        self.log.write_tty( "%s\n" %message )
                        latest = self.retrieve_latest( latest_uri )
                        for key in latest.keys():
                            if not filter.match(key):
                                del latest[key]
                        if pri == 0:
                            found_valid_site = 1
                        valid_sitelist.append(site)
                    except (DownloadException, ParseException), errmsg:
                        errmsg = "%s" % errmsg + "Skipping '%s'" %latest_uri
                        self.log.write_syslog_err( message )
                        self.log.write_stderr( errmsg ) 
                        continue
                    except:
                        raise
                    #
                    # Here we merge the parsed list with earier lists of same
                    # class:
                    #
                    self._latest_merge(main_latest, latest)
                
                if pri == 0 and not found_valid_site == 1:
                    #
                    # we cannot live without a valid main site:
                    #
                    self.log.write_syslog_err(NoValidSiteFoundError)
                    self.log.write_stderr("\n%s" %NoValidSiteFoundError)
                    sys.exit(1)
                
                latestlist.append(main_latest)

            #
            # Merge all the class-specific latest dicts into the main systems
            # latest dict. This is easy, as they are already sorted:
            #
            self.latest = {}
            for dict in latestlist:
                self.latest.update(dict)

            self.obsoletes = self._find_who_obsoletes(self.latest)
            self.obsoletors = self._find_the_obsoletors(self.latest)
            self.config["sites"] = valid_sitelist


    def regexp_filter( self, name ):
        """Filter <name> through regexps given in configuration
        file. Returns 1 if package is to be upgraded or installed, 0
        otherwise."""
        if string.lower( str(self.config["ignore_filter"]) ) == "yes":
            return True
        if self.exclude_pkg_regexp.match(name):
            return False
        elif self.include_pkg_regexp.match(name):
            return True

        
    def md5sum_isequal(self, md5sumuri, uri):
        '''Given an uri to a md5sum and an uri to the file. Check if 
        the md5sum has changed since last time the file was read.
        Record the new md5sum and return 0 if not equal and 1 if equal.'''
        old_md5s = {}
        new_md5s = {}

        md5sumfile = self.config["md5sumfile"]
        if os.path.isfile(md5sumfile):
            fd = open(md5sumfile)
            line = fd.readline()
            while line:
                split = string.split(line)
                if len(split) == 2:
                    md5, file = split
                    old_md5s[file] = md5
                line = fd.readline()
            fd.close()
        try:
            tmpfile = self.download_file(md5sumuri, CHECKSIG=0, TYPE=2)
        except DownloadException, errmsg:
            return 0
        except:
            raise
        
        fd = open(tmpfile)
        line = fd.readline()
        while line:
            split = string.split(line)
            if len(split) == 2:
                md5, file = split
                file = os.path.join(os.path.dirname(uri), file)
                new_md5s[file] = md5
            line = fd.readline()
        fd.close()

        isequal = 0
        key = uri
        if old_md5s.has_key(key) and new_md5s.has_key(key):
            if new_md5s[key] == old_md5s[key]:
                isequal = 1

        # Delete downloaded md5sum file, this lets us not flush rdfs!
        os.remove(tmpfile)
        
        return isequal

    def md5sum_set(self, md5sumuri, uri):
        '''Given an uri to a md5sum and an uri to the file. Update
        the md5sum in the file with info from the uri.'''
        old_md5s = {}
        new_md5s = {}

        md5sumfile = self.config["md5sumfile"]
        if os.path.isfile(md5sumfile):
            fd = open(md5sumfile)
            line = fd.readline()
            while line:
                split = string.split(line)
                if len(split) == 2:
                    md5, file = split
                    old_md5s[file] = md5
                line = fd.readline()
            fd.close()
        try:
            tmpfile = self.download_file(md5sumuri, CHECKSIG=0, TYPE=2)
        except DownloadException, errmsg:
            return False
        except:
            raise
        
        fd = open(tmpfile)
        line = fd.readline()
        while line:
            split = string.split(line)
            if len(split) == 2:
                md5, file = split
                file = os.path.join(os.path.dirname(uri), file)
                new_md5s[file] = md5
            line = fd.readline()
        fd.close()

        # If the md5sum has changed, store new md5sums.
        old_md5s.update(new_md5s)
        md5sumdir = os.path.dirname(md5sumfile)
        if not os.path.isdir(md5sumdir):
            os.makedirs(md5sumdir)
        fd = open(md5sumfile, 'w')
        for key in old_md5s.keys():
            md5 = old_md5s[key]
            fd.write('%s  %s\n' % (md5, key))
        fd.close()
        
        # Delete downloaded md5sum file, this lets us not flush rdfs!
        os.remove(tmpfile)
        
        return True
       
    def download_file(self, uri, siguri=None, CHECKSIG=1, TYPE=1):
        "Download file from uri and check signature. Returns local filename."
        "TYPE is: 1 = rpm-related, 2 = rdf-related"
        try:
            filename = self._downloader.download(uri, siguri, CHECKSIG, TYPE)
            return filename
        except download.AuthException, errmsg:
            raise AuthException, errmsg
        except download.DownloadException, errmsg:
            raise DownloadException, errmsg
        except:
            raise

            
    def resolve_deps(self, packages):
        '''Resolve dependencies for the list of Package-instances, return a
        resolver.transactions instance.'''
        swupgrade = False
        dep_resolver = resolver.resolver( self.reader, self.regexp_filter,
                                          self.pkgdrv, self.download_file,
                                          self.log, self.config, 
                                          self.latest, self.emptylist )
        try:
            transactions, swupgrade = dep_resolver.solve_deps( packages , 
                                                    self.obsoletors)
            info = "" 

            for transaction in transactions.values():
                for package in transaction.values():
                    info = info + \
                           "\n\t%s: %s, version %s release %s" % \
                           (package['flag'], package['name'],
                            package['version'], package['release'])
            if info != "":
                message = "Packages to install/upgrade:%s" % info
                
                # This clutters the log, we get the info from other sources.
                #self.log.write_syslog_info( message )
                self.log.write_stdout(message)
        except (resolver.DependencyError, resolver.ParseError), errmsg:
            raise ResolveException, errmsg
        except:
            raise
        return transactions, swupgrade


    def _check_package(self, flag, package, input_names):
        '''Given a package info dictionary, and the flag, return True if 
           package is accepted for install / upgrade, return False if not.
           Also return the flag, which might have been modified.'''
        if not package:
            return False, flag
        
        name = package['name']
        
        #
        # Get status of package:
        #

        status = self.pkgdrv.is_installed(package)
        #
        # Check if package is a valid upgrade:
        #
        if status == -2 and flag == UPGRADE_FLAG:
            #
            # Package is not installed, and we asked for upgrade, maybe we 
            # still 
            # need to install this package...
            #
            # Check if package obsoletes an installed package, and should
            # therefore be set to upgrade.
            #
            if package.has_key( "obsoletes" ):
                obsoletes = package["obsoletes"]
                for obs_name in obsoletes:
                    if self.pkgdrv.is_installed(obs_name) > -2:
                        message = "Package '%s' obsoletes " %name \
                                    +"installed package '%s'" \
                                    %obs_name
                        self.log.write_syslog_err(message)
                        self.log.write_stderr(message) 
                        if not self.regexp_filter(name):
                            message = "Package '%s' skipped due to " %name\
                                        +"config rule."
                            self.log.write_syslog_err(message)
                            self.log.write_stderr(message) 
                        else:
                            return True, flag
                #
                # None of the obsoletors was approved, let's give up:
                #
                if not self.emptylist:
                    message = "Package '%s' is not installed." %package['name']
                    self.log.write_syslog_err(message)
                    self.log.write_stderr("\n" +message) 
                return False, flag
            #
            # We don't want to automaticly install this package. It's up
            # to the user.
            #
            if not self.emptylist:
                message = "Package '%s' is not installed." %package['name']
                self.log.write_syslog_err(message)
                self.log.write_stderr("\n" +message) 
            return False, flag
        
        if status <= 0 and flag == UPGRADE_FLAG:
            #
            # Package is already installed with better or equal version as the 
            # potential upgrade. No need to perform upgrade.
            # 
            real_input = False
            for pkg in input_names:
                if pkg == name:
                    real_input = True
                    
            if not self.emptylist or real_input:
                message = "Package '%s' is already " %package['name']\
                            +"newest version."                 
                self.log.write_syslog_err(message)
                self.log.write_stderr(message) 
            return False, flag
            
        if status == 1 and flag == UPGRADE_FLAG:
            #
            # Installed package lesser than the potential upgrade
            # Let's check if there are more than one installed package:
            #
            
            installed = self.pkgdrv.get_packages(name)
            if len(installed) > 1:
                for instpackage in installed:
                    if self.pkgdrv.version_compare(instpackage, package) >= 0:
                        #
                        # Already installed a just as good package. Dont 
                        # perform upgrade:
                        #
                        return False, flag
            
            if not self.regexp_filter(name):
                message = "Package '%s' skipped due to config rule." %name
                self.log.write_syslog_err(message)
                self.log.write_stderr(message) 
                return False, flag

            # 
            # Now check for obsoletes, just to notify the user:
            #
            if package.has_key( "obsoletes" ):
                obsoletes = package["obsoletes"]
                for obs_name in obsoletes:
                    if self.pkgdrv.is_installed(obs_name) > -2:
                        message = "Package '%s' obsoletes " %name \
                                    +"installed package '%s'" \
                                    %obs_name
                        self.log.write_syslog_err(message)
                        self.log.write_stderr(message) 
            return True, flag
            
            
        #
        # Check if package is a valid install:
        #
        if status == -2 and flag == INSTALL_FLAG:
            #
            # Package is not installed, go ahead with installation
            #
            if not self.regexp_filter(name):
                message = "Package '%s' skipped due to config rule." %name
                self.log.write_syslog_err(message)
                self.log.write_stderr(message) 
                return False, flag
            #
            # Check for obsoletes, if found modify flag and notify user:
            #
            if package.has_key( "obsoletes" ):
                obsoletes = package["obsoletes"]
                for obs_name in obsoletes:
                    if self.pkgdrv.is_installed(obs_name) > -2:
                        message = "Package '%s' obsoletes " %name \
                                    +"installed package '%s'" \
                                    %obs_name
                        self.log.write_syslog_err(message)
                        self.log.write_stderr(message) 
                        flag = UPGRADE_FLAG
            return True, flag
            
        if status == 1 and flag == INSTALL_FLAG:
            #
            # Package is installed but lesser than potential installation,
            # should notify the user about this status, so that we can try an
            # upgrade instead. 
            #
            if not self.regexp_filter(name):
                message = "Package '%s' skipped due to config rule." %name
                self.log.write_syslog_err(message)
                self.log.write_stderr(message) 
                return False, flag
            message = "Package '%s' is installed, but potential upgrade " \
	    		%package['name'] +"was found. "  \
                        +"Upgrading."
            flag = UPGRADE_FLAG
            #
            #  Check if package obsoletes installed package and notify the user.
            #
            if package.has_key( "obsoletes" ):
                obsoletes = package["obsoletes"]
                for obs_name in obsoletes:
                    if self.pkgdrv.is_installed(obs_name) > -2:
                        message = "Package '%s' obsoletes " %name \
                                    +"installed package '%s'" \
                                    %obs_name
                        self.log.write_syslog_err(message)
                        self.log.write_stderr(message) 
            self.log.write_syslog_err(message)
            self.log.write_stderr(message) 
            return True, flag
            
        if status <= 0 and flag == INSTALL_FLAG:
            #
            # Package is installed and just as good as the potential 
            # installation.
            # No need to perform this installation:
            #
            real_input = False
            for pkg in input_names:
                if pkg == name:
                    real_input = True
                    
            if not self.emptylist or real_input:
                message = "Package '%s' is already newest version." \
                                                        %package['name']
                self.log.write_syslog_err(message)
                self.log.write_stderr(message) 
            return False, flag
        
        #
        # Don't think it's even possible to get here, but:
        #
        message = "This should never happen."
        self.log.write_syslog_err(message)
        self.log.write_stderr(message)
        return False, flag
        

    def parse_input( self, packages, flag):
        '''Given a list of package names, make sure the content is valid, 
           modify if not, and return the updated list, and information
           about noexistant packages (if needed).'''
        info = []
        none_warns = []
        package_name = None
        
        #
        # If we are going to upgrade, and we got no packages as input, 
        # lets try all:
        #
        if len(packages) == 0:
            #
            # Don't tell folks about all the newest versions... :)
            #
            self.emptylist = True
            installed = self.pkgdrv.get_installed()
            for key in installed.keys():
                if self.latest.has_key(key):
                    info.append(key)
                #
                # an installed package can be amongst the obsoleted packages
                # too, let's check:
                #
                elif self.obsoletes.has_key(key):
                    #
                    # one or more packages may obsolete it, let's try to be
                    # intelligent. If an obsoletor already is installed, we
                    # should use that one. If none is, let's take the first
                    # available.
                    #
                    picked = False
                    obsoletor = ""
                    for obsoletor in self.obsoletes[key]:
                        if installed.has_key(obsoletor):
                            info.append(obsoletor)
                            picked = True
                    if not picked:
                        obsoletor = self.obsoletes[key][0]
                        info.append(obsoletor)
                            
                    

        else:
            
            #
            # Now let's check the input given:
            #
            for name in packages:
                if self.latest.has_key(name):
                    package_name = name
                elif self.obsoletes.has_key(name):
                    #
                    # Package is obsoleted, check if one of the obsoletors is 
                    # installed or use the first one instead of the obsoleted 
                    # package:
                    #
                    message = ""
                    installed = False
                    alternative_obsoletors = []
                    #
                    # Find all possible obsoletors:
                    #
                    for obsoletor in self.obsoletors.keys():
                        for package in self.obsoletors[obsoletor]:
                            if package == name:
                                alternative_obsoletors.append(obsoletor)
                    #
                    # Try to pick one intelligently. We prefer if the 
                    # obsoletor is already installed. (Not likely but can 
                    # happen if packages have merged)
                    #
                    for obsoletor in alternative_obsoletors:
                        if self.pkgdrv.is_installed(obsoletor) == 2:
                            installed = True
                            package_name = obsoletor
                            continue
                    #
                    # No luck, pick the first available obsoletor, since we 
                    # have at least 1
                    #
                    if not installed:
                        package_name = alternative_obsoletors[0]
                else:
                    package_name = None
                    none_warns.append("Nothing found for '%s'" %name)
                #
                # Only append package name if not None
                #
                if package_name:
                    info.append(package_name)

        return info, none_warns

        


    def create_resolver_input(self, package_names, flag):
        '''Given a list of package names and a flag
        specifying "install" or "upgrade", return a tuple of input for the 
	resolver, and dict with obsoletors as keys, and a list of 
	obsoleted packages as values.'''
        
        self._generate_latest_lists()
        resolver_input = []
        siguri = None
        
        
        packages, none_warns = \
           self.parse_input(package_names, flag)
        
        #
        # User information about nonexistant packages:
        #
        if len(none_warns) > 0:
            message = ""
            for i in range( len(none_warns)):
                message += "\n" +none_warns[i]
            self.log.write_syslog_err( message )
            self.log.write_stderr( message )

        #
        # Change from package names to package dicts:
        #
        dicts = []
        for package in packages:
            if self.latest.has_key(package):
                dicts.append(self.latest[package])
        packages = dicts
        #
        # Now lets check the packages
        #
        for package in packages:
            result, newflag = self._check_package( flag, package, package_names)
            if result:
                #
                # add some needed items:
                #
                package['signature_uri'] = "%s.asc" %package['uri']
                package['flag'] = newflag
                package['resolved'] = False
                resolver_input.append( package )
        
        return resolver_input
    
    
    def install(self, transaction):
        'Given a resolver.transactions instance, run the transactions.'
        try:
            self.pkgdrv.opendb_readwrite()
            self.pkgdrv.install(transaction)
            return None
        except self.pkgdrv.install_error, errmsg:
            return errmsg
        except:
            raise


    def _main( self, packages ):
        'Common main loop for upgrade or install.'
        transactions, swupgrade = self.resolve_deps(packages)
        if not transactions:
            message = "Nothing to install or upgrade."
            self.log.write_stdout_info( message )
            self.log.write_syslog_info( message )
            return
        if string.lower( str(self.config["poll_only"]) ) == "yes":
            message = "Found these packages:\n"
            i = 0
            for transaction in transactions.values():
                for pkg in transaction.values():
                    i = i + 1
                    message = message + "\t%s: %s-%s-%s\n" %\
                          (pkg['flag'], pkg['name'], pkg['version'], 
                          pkg['release'])
            message = message + "\tTotal %i packages." %i
            self.log.write_stdout(message)
            self.log.write_syslog_info( message )
        else:
            i = 0
            errorflag = False

            for transaction in transactions.values():
                try:
                    for package in transaction.values():
                        self.log.write_tty('\rDownloading package: %s -' \
                                            % package['name'])
                        uri = utils.normalize_uri(package['uri'])
                        siguri = utils.normalize_uri(package['signature_uri'])
                        package['localfilename'] = self.download_file(uri, siguri)
                        self.log.write_tty(\
                        '\rDownloading package: %s - done\n' % package['name'])
                except DownloadException, errmsg:
                    skipped = ""
                    for name in transaction.keys():
                        skipped = skipped + name + " "
                    skipped = string.rstrip( skipped )
                    errmsg = "%s\nSkipping packages %s" % (errmsg, skipped)
                    self.log.write_syslog_err( errmsg )
                    self.log.write_stderr( errmsg )
                    del transaction[package["name"]]
                    continue
                except:
                    raise

                if self.config['save_to']:
                    todir = self.config['save_to']
                    import shutil
                    if not os.path.isdir( todir ):
                        message = 'Can not save copy of packages.'\
                                  ' No such directory'\
                                  ': %s' % todir
                        self.log.write_syslog_err( message )
                        self.log.write_stderr( message )
                    else:
                        for package in transaction.values():
                            try:
                                fromfile = package['localfilename']
                                basename = os.path.basename( fromfile )
                                tofile = os.path.join( todir, basename )
                                shutil.copyfile( fromfile, tofile )
                            except IOError:
                                message = 'Unable to move file %s to %s' % \
                                          (fromfile, tofile)
                                self.log.write_stderr( message )
                                self.log.write_syslog_err( message )
                            except:
                                raise

                if not string.lower( str(self.config["download_only"]) ) == \
                   "yes":
                    errmsg = self.install( transaction )
                    if errmsg:
                        errorflag = True
                        self.log.write_syslog_err( errmsg )
                        self.log.write_stderr( errmsg )
                i = i +1
            self.log.write_tty( "\n" )
            if errorflag:
                #self.log.write_stderr( DEPENDENCY_ERROR_MESSAGE )
                self.log.write_syslog_info( DEPENDENCY_ERROR_MESSAGE )
        return swupgrade


    def main(self):
        'Main function called by run(). Can be overloaded.'
        # Main function. Can be subclassed.
        self.swupgrade = False
        message = "Starting upgrade."
        self.log.write_syslog_info( message )
        self.log.write_stdout_verbose( message )
        selection = []
        packages = self.create_resolver_input(selection, 
                                                UPGRADE_FLAG)
        if not packages:
            message = "Your system is up to date." + 54*' '
            self.log.write_syslog_info( message )
            self.log.write_stdout_info( "\r%s" %message )
        else:
            self.swupgrade = self._main(packages)

        return self.swupgrade


    def run( self ):
        self.swupgrade = False
        try:
            self.swupgrade = self.main()
        except UpgradeException, errmsg:
            message = str(errmsg)
            self.log.write_syslog_err( message )
            self.log.write_stderr( message )
            raise
        # add handling log exc
        except log.IOException, errmsg:
            errmsg = str( errmsg )
            self.log.write_stderr( errmsg )
            self.log.write_syslog_err( errmsg ) # syslog probably works.
            raise UpgradeException, errmsg
        except:
            raise

        return self.swupgrade

                
class upgrade_package( Upgrade ):


    def __init__( self, install_list, configdict = None ):
        Upgrade.__init__( self, configdict )
        self.install_list = install_list

    
    
    def main( self ):
        message = "Starting upgrade of package."
        self.log.write_syslog_info( message )
        self.log.write_stdout_verbose(message)
        packages = self.create_resolver_input(self.install_list,
                                              UPGRADE_FLAG)
        if packages:
            self._main(packages)



    
class install_package( Upgrade ):


    def __init__( self, install_list, configdict = None ):
        Upgrade.__init__( self, configdict )
        self.install_list = install_list


    def main( self ):
        message = "Starting install."
        self.log.write_syslog_info( message )
        self.log.write_stdout_verbose( message )
        packages = self.create_resolver_input(self.install_list,
                                              INSTALL_FLAG)
        if packages:
            self._main(packages)





class list_latest( Upgrade ):


    def __init__( self, configdict = None ):
        Upgrade.__init__(self, configdict)


    def write( self ):
        self._generate_latest_lists()
        self.log.write_tty('\n')
        if not self.latest: return
        keys = self.latest.keys()
        keys.sort()
        for name in keys:
            version = self.latest[name]["version"]
            release = self.latest[name]["release"]
            uri = self.latest[name]["uri"]
            summary = self.latest[name]["summary"]
            message = "%s-%s-%s - %s" % ( name, version, release, summary )
            self.log.write_stdout( message )


    def main(self):
        self.write()



        
class list_prospect( Upgrade ):

    
    def __init__( self, configdict=None ):
        Upgrade.__init__(self, configdict)
        

    def write( self ):
        self._generate_latest_lists()
        self.log.write_tty('\n')
        if not self.latest: return
        keys = self.latest.keys()
        keys.sort()
        prospects = False
        message = ""
        for name in keys:
            if self.pkgdrv.is_installed( name ) == 2: continue
            prospects = True
            version = self.latest[name]["version"]
            release = self.latest[name]["release"]
            uri = self.latest[name]["uri"]
            summary = self.latest[name]["summary"]
            message += "\n%s-%s-%s - %s" % ( name, version, release, summary )
        if not prospects:
            message = "No prospects found."
        self.log.write_stdout( message )


    def main(self):
        self.write()


class list_alien( Upgrade ):

    def __init__(self, configdict=None ):
        Upgrade.__init__(self, configdict)

    def write ( self ):
        self._generate_latest_lists()
        if not self.latest: return
        keys = self.pkgdrv.get_installed()
        message = "The following packages are alien according "\
                  + "to the configured repositories: "
        aliens = False
        for name in keys:
            if not self.latest.has_key(name): 
                pkginfo = self.pkgdrv.get_package(name)
                message += "\n%s-%s-%s"  % (name, pkginfo['version'], 
                                                    pkginfo['release'])
                aliens = True
        if not aliens:
            message = "\nNo alien packages found."

        self.log.write_stdout( message )

    def main(self):
        self.write()


class list_downgrade( Upgrade ):

    def __init__(self, configdict=None ):
        Upgrade.__init__(self, configdict)

    def write ( self ):
        self._generate_latest_lists()
        if not self.latest: return
        keys = self.pkgdrv.get_installed()
        downgrade = False
        message = "The following installed packages have a higher " \
                    +"version number than the latest \n" \
                    +"available according to the configured repositories: "
        for package in self.latest.values():
            if self.pkgdrv.is_installed(package) == -1:
                installed = self.pkgdrv.get_package(package['name'])
                message +=  "\n%s (%s-%s vs. %s-%s)"\
                    %(installed["name"], installed['version'], 
                    installed['release'], package['version'], 
                    package['release'])
                downgrade = True
                    
        if not downgrade:
            message = "\nNo local package is newer than remote alternative."
            
        self.log.write_stdout( message )
            
    def main(self):
        self.write()


class list_upgrade( Upgrade ) :

    def __init__( self, configdict = None ):
        Upgrade.__init__(self, configdict)

    def write ( self ):
        packages = self.create_resolver_input([], UPGRADE_FLAG)
        if not packages:
            message = "Your system is up to date."
            self.log.write_syslog_info( message )
            self.log.write_stdout_info( "\n%s" %message )
        else:
            message = "The following packages will be installed if upgrade mode is chosen: "
            self.log.write_stdout_info( "\n%s" %message )
            for package in packages:
                self.log.write_stdout_info( package["name"] )
            
    def main(self):
        self.write()

class search_package( Upgrade ):
    '''Search for packages by name, given a regexp search pattern,
    print result.'''
    
    
    def __init__( self, searchpattern, configdict = None ):
        Upgrade.__init__(self, configdict)
        self.pattern = self._compile_regexp(searchpattern)
        self.matches = []
    
    
    def search( self ):
        '''Search for packages whose name matches a regexp, and write the
        output to stdout.'''
        
        self._generate_latest_lists()
        if not self.latest:
            message = 'No package information found!'
            self.log.write_stderr( message )
            self.log.write_syslog_err( message )
            return
        
        keys = self.latest.keys()
        keys.sort()
        c = 0
        message = None
        for name in keys:
            if self.pattern.search(name):
                version = self.latest[name]["version"]
                release = self.latest[name]["release"]
                uri = self.latest[name]["uri"]
                summary = self.latest[name]["summary"]
                # 
                verbose = 0
                if self.config['loglevel'] == log.LOGLEVEL_VERBOSE:
                    verbose = 1
                
                if message == None and verbose:
                    message = "%s-%s-%s - %s\n" \
                        %(name, version, release, summary)
                elif message == None and not verbose:
                    message = "%s-%s-%s\n" \
                        %(name, version, release)
                elif message and verbose:
                    message = message +  "%s-%s-%s - %s\n" \
                        %(name, version, release, summary)
                elif message and not verbose:
                    message = message + "%s-%s-%s\n" \
                        %(name, version, release)
                c = c + 1
        
        if message != None:
            sum = ( '\nFound ' + str(c) + ' package(s) matching "' +
                       self.pattern.pattern +  '":\n'  )
            self.log.write_tty(sum)
            self.log.write_stdout(string.strip(message))
        else:
            message = '\nFound 0 packages matching "' \
                            + self.pattern.pattern + '"\n'
            self.log.write_tty( message )
    
    
    def main( self ):
        self.search()




class search_file( Upgrade ):
    '''Search for packages providing file(s) matching the given searchpattern'''
    
    
    def __init__( self, searchpattern, configdict = None ):
        Upgrade.__init__(self, configdict)
        self.searchpattern = searchpattern
        self.pattern = self._compile_regexp(self.searchpattern)
        self.matches = []
    
    
    def retrieve_filelist( self, uri ):
        '''Get filelist from uri, and save. Returns a table containing
        the lines in the list.'''
        
        sitelist = self.config['sites']
        if uri[-3:] == '.gz':
            md5sum_uri = uri[:-3] + '.md5'
            noncompressed_uri = 'rdfs/'+ uri[:-3]
        else:
            noncompressed_uri = 'rdfs/'+ uri
            md5sum_uri = uri + '.md5'
        filetab = None
        
        if self.md5sum_isequal(md5sum_uri, noncompressed_uri):
            local_filelist = os.path.join(self.config['listsdir'], uri)
            local_filelist = utils.normalize_uri(local_filelist)
            if os.path.isfile(local_filelist):
                fd = open(local_filelist)
                filetab = fd.readlines()
                fd.close()

        if not filetab:
            #Try DL'ing gzipped file first.
            if uri[-3:] != '.gz':
                try:
                    local_filelist = self.download_file(uri + '.gz')
                except IOError, DownloadException:#necessary?
                    local_filelist = self.download_file(uri)
                except:
                    raise

            fd = open(local_filelist)
            filetab = fd.readlines()
            fd.close()
            # move filelist to listdir
            savefile = os.path.join(self.config['listsdir'], uri)
            savefile = utils.normalize_uri(savefile)
            os.renames(local_filelist, savefile)
            self.md5sum_set(md5sum_uri,noncompressed_uri)
        return filetab
    
    
    def get_filelists(self):
        '''Returns a dictionary containing a table of files, with
        uri as key.'''
        
        sitelist = self.config["sites"]
        sitelist.sort(self._sort_2d_sitelist)
	found_valid_site = 0
        filelists = {}
        
        for site in sitelist:
            (pri, latest_uri, regexp) = site
            list_uri = re.sub('latest.rdf', 'filelist', latest_uri)
            pri = abs(int(pri))
            try:
                message = "Fetching file info '%s'" % list_uri
                self.log.write_syslog_info( message )
                filelists[list_uri] = self.retrieve_filelist(list_uri)
		found_valid_site = 1
            except DownloadException, errmsg:
                errmsg = "%s Skipping." % errmsg
                self.log.write_syslog_err( message )
                self.log.write_stderr(errmsg)
                continue
            except:
                raise

	if not found_valid_site == 1:
            self.log.write_syslog_err(NoValidSiteFoundError)
            self.log.write_stderr("\n%s" %NoValidSiteFoundError)
            sys.exit(1)
        return filelists

    
    def search( self, listdict ):
        '''Search for packages providing files that matches the
        searchpattern. Prints output to stdout'''
        
        matched_at_all = False
        header = ""
        message = ""
        for uri in listdict.keys():
            header = 'from ' + uri + ':'
            found = False
            for line in listdict[uri]:
                if line[0] == "[":
                    package = line[10:-15]
                    continue
                filepath, filename  = os.path.split(line)
	        matched = self.pattern.search(filename)
                if matched:
                    found = True
		    message += "%s: %s" %(package,line)
            if found:
                matched_at_all = True
                self.log.write_tty(header+message)
        if not matched_at_all:
            message = "Found 0 files matching \"%s\"\n" %self.searchpattern
            self.log.write_tty(message)
    
    
    def main( self ):
        
        filelists = self.get_filelists()
        if not filelists:
            self.log.write_stderr('No file information found!')
            return
        
        self.search(filelists)

class describe( Upgrade ):
    '''show description for the package given as argument'''
    
    def __init__( self, configdict = None ):
        Upgrade.__init__(self, configdict)
        self.target = configdict["descriptor"]
        self.summary = None
    
    
    def find_package( self ):
        '''Find package that fits descriptor'''
        
        self._generate_latest_lists()
        
        if not self.latest:
            message = 'No package information found!'
            self.log.write_stderr( "" )
            self.log.write_stderr( message )
            self.log.write_syslog_err( message )
            return
        
        pkg = {}
        if self.latest.has_key(self.target):
            pkg = self.latest[self.target]
        else:
            message = "No package named '%s' found!" %self.target
            self.log.write_stderr( "" )
            self.log.write_stderr( message)
            self.log.write_syslog_err( message )
            return

        uri = pkg['uri']
        try:
            (baseuri, basename) = os.path.split(uri)
            self.log.write_tty('Parsing file: %s -' % basename)
            local_filename = self.download_file( uri,TYPE=2 )
            package = self.reader.read_package( local_filename, baseuri )
            self.log.write_tty('\rParsing file: %s - done' % basename)
        except spiparser.parse_error:
            raise ParseException, "Unable to parse file '%s'" % uri
        except:
            raise


        result = "\n"
        
        if package.has_key("name"):
            result += "Name: %s\n" %package["name"]
        if package.has_key("epoch"):
            result += "Epoch: %s\n" %package["epoch"]
        if package.has_key("serial"):
            result += "Serial: %s\n" %package["serial"]
        if package.has_key("version"):
            result += "Version: %s\n" %package["version"]
        if package.has_key("release"):
            result += "Release: %s\n" %package["release"]
        if package.has_key("distribution"):
            result += "Distribution: %s\n" %package["distribution"]
        if package.has_key("vendor"):
            result += "Vendor: %s\n" %package["vendor"]
        if package.has_key("copyright"):
            result += "License: %s\n" %package["copyright"]
        if package.has_key("group"):
            result += "Group: %s\n" %package["group"]
        if package.has_key("build_date"):
            result += "Build date: %s\n" %package["build_date"]
        if package.has_key("arch"):
            result += "Arch: %s\n" %package["arch"]
        if package.has_key("os"):
            result += "OS: %s\n" %package["os"]
        if package.has_key("size"):
            result += "Size: %s\n" %package["size"]
        if package.has_key("md5sum"):
            result += "Md5sum: %s\n" %package["md5sum"]
        if package.has_key("uri"):
            result += "Uri: %s\n" %package["uri"]
        if package.has_key("signature"):
            result += "Sign: %s\n" %package["signature"]
        if package.has_key("requirements"):
            requires = package["requirements"]
            reqmess = ""
            index = 0
            handled = {}
            for req in requires:
                name = req["name"]
                if handled.has_key(name):
                    continue
                if index > 0:
                    reqmess =  reqmess + ", "
                if req['operator'] == '':
                    reqmess = reqmess + "%s" %name
                else:
                    reqmess = reqmess + "%s " %name + \
                        "(%s " %req['operator'] + "%s-" %req['version'] +\
                        "%s)" %req['release']
                index += 1

                handled[name] = 1
            if not reqmess == "":
                result += "Requires: %s\n" %reqmess
        if package.has_key("conflicts"):
            conflicts = package["conflicts"]
            conmess = ""
            index = 0
            for con in conflicts:
                if index > 0:
                    conmess =  conmess + ", "
                conmess = conmess + "%s " %con
                index += 1

            if not conmess == "":
                result += "Conflicts: %s\n" %conmess
        if package.has_key("summary"):
            result += "\nSummary: %s\n" %package["summary"]
        if package.has_key("description"):
            result += "\nDescription: %s\n" %package["description"]


            
        self.log.write_stdout( result )
    
        return
    
    def main( self ):
        self.find_package()



